import torch

class Callback:
    def __init__(self):
        pass

### Events
    def on_start(self) -> None:
        pass

    def on_finish(self) -> None:
        pass

    def on_epoch_start(self, epoch: int, model: 'Model') -> None:
        pass

    def on_epoch_finish(self, epoch: int, model: 'Model') -> None:
        pass

    def on_batch_start(self, batch, batch_id: int, epoch: int) -> None:
        pass

    def on_batch_finish(self, batch, batch_id: int, epoch: int,
                        loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        pass

    def on_training_epoch_start(self, epoch: int, model: 'Model') -> None:
        pass

    def on_training_epoch_finish(self, epoch: int, model: 'Model') -> None:
        pass

    def on_validation_epoch_start(self, epoch: int, model: 'Model') -> None:
        pass

    def on_validation_epoch_finish(self, epoch: int, model: 'Model') -> None:
        pass

    def on_testing_start(self) -> None:
        pass

    def on_testing_finish(self) -> None:
        pass

    def on_training_batch_start(self, batch, batch_id: int, epoch: int) -> None:
        pass

    def on_training_batch_finish(self, batch, batch_id: int, epoch: int,
                                 loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        pass

    def on_validation_batch_start(self, batch, batch_id: int, epoch: int) -> None:
        pass

    def on_validation_batch_finish(self, batch, batch_id: int, epoch: int,
                                   loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        pass

    def on_testing_batch_start(self, batch, batch_id: int) -> None:
        pass

    def on_testing_batch_finish(self, batch, batch_id: int,
                                loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        pass



from ..core import Callback
import pandas as pd
import numpy as np
from pathlib import Path
from typing import Tuple

class History(Callback):
    def __init__(self):
        super(History, self).__init__()
        self.history = []

    def on_epoch_finish(self, epoch: int, model: 'Model') -> None:
        epoch_metrics = {m.name: m.value for m in model.metrics}
        self.history.append(epoch_metrics)


class TensorboardCallback(Callback):
    def __init__(self, log_dir: str = None, flush_secs: int = 30):
        super(TensorboardCallback, self).__init__()
        try:
            from torch.utils.tensorboard import SummaryWriter
        except ImportError:
            try:
                from tensorboardX import SummaryWriter
            except ImportError:
                raise ValueError('No Tensorboard installed, can\'t use TensorboardCallback')

        self.tb = SummaryWriter(flush_secs=flush_secs, log_dir=log_dir)

    def on_epoch_finish(self, epoch: int, model: 'Model') -> None:
        super(TensorboardCallback, self).on_epoch_finish(epoch, model)
        for m in model.metrics:
            if isinstance(m.value, list):
                self.tb.add_histogram(m.name, np.array(m.value), epoch)
            else:
                self.tb.add_scalar(m.name, m.value, epoch)
        self.tb.flush()


class SpreadsheetCallback(Callback):
    def __init__(self, path: Path, file_name: str = 'history.xlsx', float_format: str = '%.8f',
                 freeze_panes: Tuple = (1, 1), sheet: str = 'Sheet1'):
        super(SpreadsheetCallback, self).__init__()

        # dataframe to store metrics
        self.history = pd.DataFrame()

        # save parameters
        self.sheet_name = sheet
        self.float_format = float_format
        self.freeze_panes = freeze_panes
        self.file_name = file_name
        self.path = path

    def on_epoch_finish(self, epoch: int, model: 'Model') -> None:
        super(SpreadsheetCallback, self).on_epoch_finish(epoch, model)

        epoch_metrics = {m.name: m.value for m in model.metrics}
        self.history.append(epoch_metrics)

        self.history.to_excel(self.path / self.file_name, sheet_name=self.sheet_name, float_format=self.float_format,
                              freeze_panes=self.freeze_panes)
        print('TEST')



import torch
import torch.nn as nn
from torch.nn import init
import torch.nn.functional as F
from typing import Union, List

### General blocks
def init_weights(net: nn.Module, init_type: str = 'normal', init_gain: float = 0.02):
    # https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/blob/master/models/networks.py
    """Initialize network weights.
    Parameters:
        net (network)   -- network to be initialized
        init_type (str) -- the name of an initialization method: normal | xavier | kaiming | orthogonal
        init_gain (float)    -- scaling factor for normal, xavier and orthogonal.
    """
    def init_func(m):  # define the initialization function
        classname = m.__class__.__name__
        if hasattr(m, 'weight') and (classname.find('Conv') != -1 or classname.find('Linear') != -1):
            if init_type == 'normal':
                init.normal_(m.weight.data, 0.0, init_gain)
            elif init_type == 'xavier':
                init.xavier_normal_(m.weight.data, gain=init_gain)
            elif init_type == 'kaiming':
                init.kaiming_normal_(m.weight.data, a=0, mode='fan_in')
            elif init_type == 'orthogonal':
                init.orthogonal_(m.weight.data, gain=init_gain)
            else:
                raise NotImplementedError('initialization method [%s] is not implemented' % init_type)
            if hasattr(m, 'bias') and m.bias is not None:
                init.constant_(m.bias.data, 0.0)
        elif classname.find('BatchNorm') != -1:
            init.normal_(m.weight.data, 1.0)
            init.constant_(m.bias.data, 0.0)

    net.apply(init_func)

#### 3D U-Net building blocks
class ReplicateConv3d(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: Union[int, List[int]] = 3, padding: int = 1,
                 bias: bool = False):
        super(ReplicateConv3d, self).__init__()

        self.conv = nn.Conv3d(in_channels, out_channels, kernel_size=kernel_size, padding=0, bias=bias)
        self.padding = padding

    def forward(self, x):
        x = F.pad(x, [self.padding, self.padding, self.padding, self.padding, self.padding, self.padding], mode='replicate')
        x = self.conv(x)
        return x


class ConvBnActivation3D(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int = 3, padding: int = -1,
                 bn: bool = True,
                 activation: str = 'leakyrelu', activation_inplace: bool = True):
        super(ConvBnActivation3D, self).__init__()

        self.padding = (kernel_size - 1) // 2
        if padding == -1:
            self.conv = ReplicateConv3d(in_channels, out_channels, kernel_size=kernel_size, padding=self.padding, bias=False)
        else:
            self.conv = nn.Conv3d(in_channels, out_channels, kernel_size=kernel_size, padding=self.padding, bias=False)

        if bn:
            self.bn = nn.BatchNorm3d(out_channels)
        else:
            self.bn = None

        if activation == 'relu':
            self.activation = nn.ReLU(inplace=activation_inplace)
        elif activation == 'leakyrelu' or activation == 'leaky_relu' or activation == 'lrelu':
            self.activation = nn.LeakyReLU(inplace=activation_inplace)
        elif activation == 'selu':
            self.activation = nn.SELU(inplace=activation_inplace)
        elif activation == 'elu':
            self.activation = nn.ELU(inplace=activation_inplace)
        else:
            raise NotImplementedError('Check ConvBnActivation class for available activations')

    def forward(self, x):
        x = self.conv(x)
        if self.bn:
            x = self.bn(x)
        x = self.activation(x)
        return x

class EncoderBlock3D(nn.Module):

    def __init__(self, x_channels: int, y_channels: int, kernel_size: int = 3, residual: bool = False,
                 bn: bool = True,
                 activation: str = 'leakyrelu', activation_inplace: bool = True,
                 pooling: str = 'max'):
        super(EncoderBlock3D, self).__init__()

        self.padding = (kernel_size - 1) // 2
        self.residual = residual

        self.encode = nn.Sequential(
            ConvBnActivation3D(x_channels, y_channels, kernel_size=kernel_size, padding=self.padding, bn=bn,
                               activation=activation, activation_inplace=activation_inplace),
            ConvBnActivation3D(y_channels, y_channels, kernel_size=kernel_size, padding=self.padding, bn=bn,
                               activation=activation, activation_inplace=activation_inplace),
        )
        if self.residual:
            self.encode2 = ConvBnActivation3D(y_channels + x_channels, y_channels, kernel_size=kernel_size,
                                              padding=self.padding, bn=bn, activation=activation,
                                              activation_inplace=activation_inplace)

        if pooling == 'max':
            self.pool = nn.MaxPool3d(kernel_size=2, stride=2)
        elif pooling == 'avg':
            self.pool = nn.AvgPool3d(kernel_size=2, stride=2)
        else:
            raise NotImplementedError('Check EncoderBlock3D class for available pooling operations')

    def forward(self, x: torch.Tensor):
        input_tensor = x
        x = self.encode(x)
        if self.residual:
            x = torch.cat([x, input_tensor], dim=1)
            x = self.encode2(x)
        x = self.pool(x)
        return x

class DecoderBlock3D(nn.Module):
    def __init__(self, x_external_channels: int, x_channels: int, y_channels: int, kernel_size: int = 3,
                 bn: bool = True,
                 activation: str = 'leakyrelu', activation_inplace: bool = True,
                 upsampling: str = 'trilinear'):
        super(DecoderBlock3D, self).__init__()

        padding = (kernel_size - 1) // 2

        self.up_0 = ConvBnActivation3D(x_external_channels + x_channels, x_channels, kernel_size=kernel_size, padding=padding,
                                       bn=bn, activation=activation, activation_inplace=activation_inplace)
        self.up_1 = ConvBnActivation3D(x_channels, y_channels, kernel_size=kernel_size, padding=padding,
                                       bn=bn, activation=activation, activation_inplace=activation_inplace)
        self.upsample = nn.Upsample(scale_factor=2, mode=upsampling, align_corners=True)
        self.up_2 = ConvBnActivation3D(y_channels, y_channels, kernel_size=kernel_size, padding=padding,
                                       bn=bn, activation=activation, activation_inplace=activation_inplace)

    def forward(self, x: torch.Tensor, down_tensor: torch.Tensor):
        x = torch.cat([x, down_tensor], 1)
        x = self.up_0(x)
        skip = x
        x = self.up_1(x)
        x = self.upsample(x)
        x = self.up_2(x)

        return x, skip



from typing import List, Any, Union
import torch
import torch.nn.functional as F
from kinlin.core import Metric, RunningEpochMetric
from kinlin.core.utils import to_onehot
import numpy as np

### Classwise metrics

# Tversky index: https://en.wikipedia.org/wiki/Tversky_index
# a = b = 0.5 = dice coefficient
# a = b = 1 = jaccard index / tanimoto coefficient
class ClasswiseTverskyIndex(RunningEpochMetric):
    def __init__(self, n_classes: int, alpha: float = 0.5, beta: float = 0.5, eps: float = 1e-7):
        super(ClasswiseTverskyIndex, self).__init__()
        self.n_classes = n_classes
        self.alpha = alpha
        self.beta = beta
        self.eps = eps

    def print(self) -> str:
        return f'[{",".join(f"{k:.3f}" for k in self.value)}]'

    def on_batch_finish(self, batch, batch_id: int, epoch: int, loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        super(ClasswiseTverskyIndex, self).on_batch_finish(batch, batch_id, epoch, loss, y_pred, y_true)

        y_true_onehot = to_onehot(y_true, self.n_classes).float()
        prob = F.softmax(y_pred, dim=1)

        dims = (0, 2, 3, 4)
        intersection = torch.sum(prob * y_true_onehot, dims)
        fps = torch.sum(prob * torch.sub(1, y_true_onehot), dims)
        fns = torch.sum(torch.sub(1, prob) * y_true_onehot, dims)

        del y_true_onehot, prob

        num = intersection
        den = intersection + (self.alpha * fps) + (self.beta * fns)
        tversky_index = (num / (den + self.eps))

        self.set(tversky_index.cpu().numpy())

class ClasswiseDiceCoefficient(ClasswiseTverskyIndex):
    def __init__(self, n_classes: int):
        super(ClasswiseDiceCoefficient, self).__init__(n_classes, alpha=0.5, beta=0.5)

class ClasswiseJaccardIndex(ClasswiseTverskyIndex):
    def __init__(self, n_classes: int):
        super(ClasswiseJaccardIndex, self).__init__(n_classes, alpha=1.0, beta=1.0)

class ClasswiseTanimotoCoefficient(ClasswiseJaccardIndex):
    # alias for Jaccard index
    pass


### Averaged metrics
class TverskyIndex(ClasswiseTverskyIndex):
    def on_batch_finish(self, batch, batch_id: int, epoch: int, loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        super(TverskyIndex, self).on_batch_finish(batch, batch_id, epoch, loss, y_pred, y_true)
        self.value = self.value.mean()

    def print(self) -> str:
        return f'{self.value:.3f}'

class DiceCoefficient(TverskyIndex):
    def __init__(self, n_classes: int):
        super(DiceCoefficient, self).__init__(n_classes, alpha=0.5, beta=0.5)

class JaccardIndex(TverskyIndex):
    def __init__(self, n_classes: int):
        super(JaccardIndex, self).__init__(n_classes, alpha=1.0, beta=1.0)

class TanimotoCoefficient(JaccardIndex):
    # alias for Jaccard index
    pass



import torch
import numpy as np
from typing import Union, List, Callable, Tuple
from .utils import generate_random_id, readable_number
from .metric import Metric, Epoch, SemSegAccuracy, Loss, EpochTimer

class Model:

    def __init__(self, network: torch.nn.Module, name: str = None, metrics: List[Metric] = None):

        # properties
        self.id: str = generate_random_id()
        self.name: str = name if name else self.network.__class__.__name__
        self.network: torch.nn.Module = network
        self.metrics = metrics if metrics else []

        self.network.to(device='cuda')

        # set default metrics
        self.metrics.append(Epoch())
        self.metrics.append(Loss())
        self.metrics.append(SemSegAccuracy())
        self.metrics.append(EpochTimer())

### Methods called by training strategies
    def training_fn(self, batch, batch_id: int, epoch: int)\
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Method should describe training process on one batch and return loss, y_pred and y_true
        :param batch:
        :param batch_id:
        :param epoch:
        :return: loss, y_pred, y_true
        """
        # must return loss to backprop
        raise NotImplementedError

    def backprop_fn(self, loss: torch.Tensor, optimizer: torch.optim.Optimizer) -> None:
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

    def validation_fn(self, batch, batch_id: int, epoch: int)\
            -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """

        :param batch:
        :param batch_id:
        :param epoch:
        :return: loss, y_pred, y_true
        """
        raise NotImplementedError

    def testing_fn(self, batch: torch.Tensor, batch_id: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        raise NotImplementedError

### General methods
    def nparams(self, readable: bool = False, trainable: bool = True) -> Union[int, str]:
        if trainable:
            params = sum(p.numel() for p in self.network.parameters() if p.requires_grad)
        else:
            params = sum(p.numel() for p in self.network.parameters())

        if readable:
            return readable_number(params, unit='', decimal=3)
        else:
            return params

    def __repr__(self):
        summary = f'Model: {self.name}\n' \
                  f'\tTrainable parameters: {self.nparams(readable=True)}\n' \
                  f'\tMetrics: {", ".join(m.name for m in self.metrics)}'

        return summary

    def print_summary(self) -> None:
        print(repr(self))

    def progressbar_metrics(self, skip_metrics: List[str] = None) -> dict:
        if skip_metrics is None:
            skip_metrics = []

        metrics = {k.name: k.print() for k in self.metrics
                   if not isinstance(k.value, np.ndarray) and k.name not in skip_metrics and k.visible_progress}
        return metrics

### Events
    def on_start(self) -> None:
        for m in self.metrics:
            m.on_start()

    def on_finish(self) -> None:
        for m in self.metrics:
            m.on_finish()

    def on_epoch_start(self, epoch: int, model: 'Model') -> None:
        for m in self.metrics:
            m.on_epoch_start(epoch, model)

    def on_epoch_finish(self, epoch: int, model: 'Model') -> None:
        for m in self.metrics:
            m.on_epoch_finish(epoch, model)

    def on_batch_start(self, batch, batch_id: int, epoch: int) -> None:
        for m in self.metrics:
            m.on_batch_start(batch, batch_id, epoch)

    def on_batch_finish(self, batch, batch_id: int, epoch: int,
                        loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        for m in self.metrics:
            m.on_batch_finish(batch, batch_id, epoch, loss, y_pred, y_true)

    def on_training_epoch_start(self, epoch: int, model: 'Model') -> None:
        for m in self.metrics:
            m.on_training_epoch_start(epoch, model)
        self.on_epoch_start(epoch, model)

    def on_training_epoch_finish(self, epoch: int, model: 'Model') -> None:
        for m in self.metrics:
            m.on_training_epoch_finish(epoch, model)
        self.on_epoch_finish(epoch, model)

    def on_validation_epoch_start(self, epoch: int, model: 'Model') -> None:
        for m in self.metrics:
            m.on_validation_epoch_start(epoch, model)
        self.on_epoch_start(epoch, model)

    def on_validation_epoch_finish(self, epoch: int, model: 'Model') -> None:
        for m in self.metrics:
            m.on_validation_epoch_finish(epoch, model)
        self.on_epoch_finish(epoch, model)

    def on_testing_start(self) -> None:
        for m in self.metrics:
            m.on_testing_start()
        self.on_epoch_start(-1, self)

    def on_testing_finish(self) -> None:
        for m in self.metrics:
            m.on_testing_finish()
        self.on_epoch_finish(-1, self)

    def on_training_batch_start(self, batch, batch_id: int, epoch: int) -> None:
        for m in self.metrics:
            m.on_training_batch_start(batch, batch_id, epoch)
        self.on_batch_start(batch, batch_id, epoch)

    def on_training_batch_finish(self, batch, batch_id: int, epoch: int,
                                 loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        for m in self.metrics:
            m.on_training_batch_finish(batch, batch_id, epoch, loss, y_pred, y_true)
        self.on_batch_finish(batch, batch_id, epoch, loss, y_pred, y_true)

    def on_validation_batch_start(self, batch, batch_id: int, epoch: int) -> None:
        for m in self.metrics:
            m.on_validation_batch_start(batch, batch_id, epoch)
        self.on_batch_start(batch, batch_id, epoch)

    def on_validation_batch_finish(self, batch, batch_id: int, epoch: int,
                                   loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        for m in self.metrics:
            m.on_validation_batch_finish(batch, batch_id, epoch, loss, y_pred, y_true)
        self.on_batch_finish(batch, batch_id, epoch, loss, y_pred, y_true)

    def on_testing_batch_start(self, batch, batch_id: int) -> None:
        for m in self.metrics:
            m.on_testing_batch_start(batch, batch_id)
        self.on_batch_start(batch, batch_id, -1)

    def on_testing_batch_finish(self, batch, batch_id: int,
                                loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        for m in self.metrics:
            m.on_testing_batch_finish(batch, batch_id, loss, y_pred, y_true)
        self.on_batch_finish(batch, batch_id, -1, loss, y_pred, y_true)



from pathlib import Path
import torch
from torch.utils.data.dataset import Dataset as TorchDataset
from torch.utils.data import random_split, DataLoader
from collections import OrderedDict
import mrcfile as mrc
import numpy as np
np.set_printoptions(precision=2)
import warnings
import json
from tqdm import tqdm
warnings.filterwarnings("ignore")
import random
from kinlin import Dataset

class SHRECDataset(TorchDataset):

    def __init__(self, path: Path, mrcs: list, patch_size: int, augmentation: bool = True,
                 full_tomo: bool = True, classmasks: bool = True, grandmodel: bool = False, even_odd: bool = False):
        # inputs
        self.path = path
        self.mrcs = mrcs
        self.patch_size = patch_size
        self.augmentation = augmentation

        # find tomograms
        self.tomo_dirs = [f for f in self.path.iterdir() if f.is_dir() and f.name.isdigit() and int(f.name) in mrcs]
        self.tomo_dirs.sort()
        self.tomo_count = len(self.tomo_dirs)

        # loading progressbar settings
        total = full_tomo * self.tomo_count + classmasks * self.tomo_count + grandmodel * self.tomo_count \
                + even_odd * 2 * self.tomo_count

        t = tqdm(total=total, desc=f'Loading data', unit='mrc', ncols=100, mininterval=0.5)

        # reconstructions
        if full_tomo:
            self.reconstructions = []

            for f in self.tomo_dirs:
                with mrc.open(f / f'cropped_reconstruction_{f.name}.mrc', mode='r', permissive=True) as m:
                    tomo_data = m.data.copy().astype(np.float32)
                    tomo_data = (tomo_data - np.mean(tomo_data)) / np.std(tomo_data)
                    tomo_data = (tomo_data - np.min(tomo_data)) / (np.max(tomo_data) - np.min(tomo_data))
                    self.reconstructions.append(tomo_data)
                t.update()
        else:
            self.reconstructions = None

        # classmasks
        if classmasks:
            self.classmasks = []

            for f in self.tomo_dirs:
                with mrc.open(f / f'dilated_classmask_{f.name}.mrc', mode='r', permissive=True) as m:
                    tomo_data = m.data.copy().astype(np.int64)
                    self.classmasks.append(tomo_data)
                t.update()
        else:
            self.classmasks = None

        # even and odd reconstructions
        if even_odd:
            self.evens = []
            self.odds = []

            for f in self.tomo_dirs:
                with mrc.open(f / f'cropped_new_reconstruction_{f.name}_even.mrc', mode='r', permissive=True) as m:
                    tomo_data = m.data.copy().astype(np.float32)
                    tomo_data = (tomo_data - np.mean(tomo_data)) / np.std(tomo_data)
                    tomo_data = (tomo_data - np.min(tomo_data)) / (np.max(tomo_data) - np.min(tomo_data))
                    self.evens.append(tomo_data)
                t.update()

            for f in self.tomo_dirs:
                with mrc.open(f / f'dilated_classmask_{f.name}_odd.mrc', mode='r', permissive=True) as m:
                    tomo_data = m.data.copy().astype(np.float32)
                    tomo_data = (tomo_data - np.mean(tomo_data)) / np.std(tomo_data)
                    tomo_data = (tomo_data - np.min(tomo_data)) / (np.max(tomo_data) - np.min(tomo_data))
                    self.odds.append(tomo_data)
                t.update()
        else:
            self.evens = None
            self.odds = None

        # ground truth
        if grandmodel:
            self.gts = []
            for f in self.tomo_dirs:
                with mrc.open(f / f'grandmodel_{f.name}.mrc', mode='r', permissive=True) as m:
                    tomo_data = m.data.copy().astype(np.float32)
                    tomo_data = (tomo_data - np.mean(tomo_data)) / np.std(tomo_data)
                    tomo_data = (tomo_data - np.min(tomo_data)) / (np.max(tomo_data) - np.min(tomo_data))
                    self.gts.append(tomo_data)
                t.update()
        else:
            self.gts = None
        t.close()

        # load glossaries
        with open(str(self.path / 'num2pdb.json'), 'r') as fp:
            self.num2pdb = json.load(fp)
            # json dict keys are always strings, convert them to ints
            self.num2pdb = OrderedDict({int(k): v for k, v in self.num2pdb.items()})
        with open(str(self.path / 'pdb2num.json'), 'r') as fp:
            self.pdb2num = json.load(fp)

        # calculate classes count
        self.n_cl = len(self.num2pdb.keys())

        # tomograms shape
        try:
            self.tomo_dim = self.reconstructions[0].shape
        except:
            self.tomo_dim = self.classmasks[0].shape

        # patches calculation
        self.patch_side = self.patch_size // 2
        self.patch_overlap = self.patch_size // 3

        self.step = self.patch_size + 1 - self.patch_overlap
        self.p_centerZ = list(range(self.patch_side, self.tomo_dim[0] - self.patch_side, self.step))
        self.p_centerY = list(range(self.patch_side, self.tomo_dim[1] - self.patch_side, self.step))
        self.p_centerX = list(range(self.patch_side, self.tomo_dim[2] - self.patch_side, self.step))

        self.i_patches = np.array(np.meshgrid(list(range(self.tomo_count)),
                                              self.p_centerZ, self.p_centerY, self.p_centerX)).T.reshape(-1, 4)
        self.n_patches = self.i_patches.shape[0]

        t.write(f'Loaded {self.tomo_count} tomograms, split into {self.n_patches} patches of {self.patch_size}^3 voxels.')
        t.write(f'Each tomogram has: {"reconstructions," if full_tomo else ""} {"classmasks," if classmasks else ""}'
                f'{"ground truth," if grandmodel else ""} {"even and odd reconstructions" if even_odd else ""}\n')

    def __len__(self):
        return self.n_patches

    def __sample(self, item):

        # sample configuration
        tid = self.i_patches[item][0]  # tomo id
        center_position = self.i_patches[item][1:4]
        dim = list(zip(np.maximum(center_position - self.patch_side, 0),
                       np.minimum(center_position + self.patch_side, self.reconstructions[tid].shape)))

        # sanity checks, should never happen
        if dim[0][1] - dim[0][0] < self.patch_size or \
           dim[1][1] - dim[1][0] < self.patch_size or \
           dim[2][1] - dim[2][0] < self.patch_size:
            raise IOError

        sample = {
            'reconstruction': self.reconstructions[tid][dim[0][0]:dim[0][1], dim[1][0]:dim[1][1], dim[2][0]:dim[2][1]].copy() if self.reconstructions else None,
            'classmask': self.classmasks[tid][dim[0][0]:dim[0][1], dim[1][0]:dim[1][1], dim[2][0]:dim[2][1]].copy() if self.classmasks else None,
            'odd_reconstruction': self.odds[tid][dim[0][0]:dim[0][1], dim[1][0]:dim[1][1], dim[2][0]:dim[2][1]].copy() if self.odds else None,
            'even_reconstruction': self.evens[tid][dim[0][0]:dim[0][1], dim[1][0]:dim[1][1], dim[2][0]:dim[2][1]].copy() if self.evens else None,
            'gt': self.gts[tid][dim[0][0]:dim[0][1], dim[1][0]:dim[1][1], dim[2][0]:dim[2][1]].copy() if self.gts else None
        }

        return sample

    def __augment(self, sample):

        # fliplr
        if np.random.uniform() < 0.5:
            for k in sample:
                if sample[k] is not None:
                    sample[k] = np.fliplr(sample[k]).copy()  # fliplr returns view

        # rot 180
        if np.random.uniform() > 0.5:
            for k in sample:
                if sample[k] is not None:
                    rotation = 2  # 180 deg
                    axes = (0, 2)  # tilt axis only?
                    sample[k] = np.rot90(sample[k], k=rotation, axes=axes).copy()  # rot90 returns view

        return sample

    def __expand_and_filter(self, sample):
        # add channel dimension in front
        add_dimension = ['reconstruction', 'odd_reconstruction', 'even_reconstruction']
        for k in sample:
            if k in add_dimension and sample[k] is not None:
                sample[k] = sample[k][np.newaxis, ...]

        # filter empty samples
        sample = {k: v for k, v in sample.items() if v is not None}

        return sample

    def __getitem__(self, item):  # 0 to self.n_patches

        # sample appropriate patches from tomograms and additional files
        sample = self.__sample(item)

        # import matplotlib.pyplot as plt
        # fig, ax = plt.subplots(2, 2)
        # ax[0, 0].imshow(sample['reconstruction'][32])
        # ax[1, 0].imshow(sample['classmask'][32])

        # augment samples
        if self.augmentation:
            sample = self.__augment(sample)

        # ax[0, 1].imshow(sample['reconstruction'][32])
        # ax[1, 1].imshow(sample['classmask'][32])
        # plt.show()

        # return expanded (added channel dimension for appropriate samples) and filtered (removed None samples) samples
        return self.__expand_and_filter(sample)

class SHREC(Dataset):
    def __init__(self, path_to_shrec: Path, subtomo_size, augmentation: bool):
        super(SHREC, self).__init__()

        self.train_dataset = SHRECDataset(path_to_shrec, list(range(7, 8)), subtomo_size, augmentation=augmentation)
        self.validation_dataset = SHRECDataset(path_to_shrec, [8], subtomo_size, augmentation=augmentation)
        self.test_dataset = SHRECDataset(path_to_shrec, [9], subtomo_size, augmentation=False)

    def training_dataloader(self, batch_size=1, shuffle=True):
        return DataLoader(self.train_dataset, batch_size=batch_size, shuffle=shuffle, pin_memory=True, drop_last=True)

    def validation_dataloader(self, batch_size=1, shuffle=True):
        return DataLoader(self.validation_dataset, batch_size=batch_size, shuffle=shuffle, pin_memory=True, drop_last=True)

    def test_dataloader(self, batch_size=1, shuffle=True):
        return DataLoader(self.test_dataset, batch_size=batch_size, shuffle=True, pin_memory=True, drop_last=True)


from .overlap import ClasswiseTverskyIndex, TverskyIndex
from .overlap import ClasswiseDiceCoefficient, DiceCoefficient
from .overlap import ClasswiseJaccardIndex, JaccardIndex
from .overlap import ClasswiseTanimotoCoefficient, TanimotoCoefficient



import torch
import torch.nn as nn
import numpy as np

from enum import Enum
from typing import List, Callable, Any
from tqdm import tqdm

from .model import Model
from .dataset import Dataset
from .experiment import Experiment
from .callback import Callback

class TrainingEvents(Enum):
    START = 'on_start'
    FINISH = 'on_finish'
    TRAINING_EPOCH_START = 'on_training_epoch_start'
    TRAINING_EPOCH_FINISH = 'on_training_epoch_finish'
    TRAINING_BATCH_START = 'on_training_batch_start'
    TRAINING_BATCH_FINISH = 'on_training_batch_finish'
    VALIDATION_EPOCH_START = 'on_validation_epoch_start'
    VALIDATION_EPOCH_FINISH = 'on_validation_epoch_finish'
    VALIDATION_BATCH_START = 'on_validation_batch_start'
    VALIDATION_BATCH_FINISH = 'on_validation_batch_finish'
    TESTING_START = 'on_testing_start'
    TESTING_FINISH = 'on_testing_finish'
    TESTING_BATCH_START = 'on_testing_batch_start'
    TESTING_BATCH_FINISH = 'on_testing_batch_finish'

class TrainingStrategy:
    def __init__(self, model: Model, dataset: Dataset, optimizer: torch.optim.Optimizer, experiment: Experiment = None,
                 callbacks: List[Callback] = None):

        # properties
        self.model: Model = model
        self.dataset: Dataset = dataset
        self.optimizer: torch.optim.Optimizer = optimizer
        #self.experiment: Experiment = experiment
        self.callbacks: List[Callback] = callbacks

        # parallelize network depending on experiment settings
        # if len(self.experiment.devices) > 1:
        #     self.network = nn.DataParallel(self.model.network, device_ids=self.experiment.devices)
        # else:
        self.network = self.model.network

        # event handler
        self.handlers = {k: [] for k in TrainingEvents}

        # register events
        for event in TrainingEvents:
            # model events
            self.on_event(event, getattr(self.model, event.value))

            # callback events
            for c in self.callbacks:
                self.on_event(event, getattr(c, event.value))

    def on_event(self, event: TrainingEvents, handler: Callable):
        self.handlers[event].append(handler)

    def emit(self, event: TrainingEvents, *args, **kwargs):
        for handler in self.handlers[event]:
            handler(*args, **kwargs)

    def training_epoch(self, epoch: int) -> None:
        raise NotImplementedError

    def validation_epoch(self, epoch: int) -> None:
        raise NotImplementedError

    def test(self) -> None:
        raise NotImplementedError

    def __call__(self, n_epochs: int = 1, validation: bool = True, verbose: bool = True):
        if verbose:
            print(f'Training{" and validating" if validation else ""}'
                  f' for {n_epochs} {"epochs" if n_epochs > 1 else "epoch"}')
            self.model.print_summary()
            self.dataset.print_summary()
            print(f'Optimizer: {self.optimizer.__class__.__name__}\n'
                  f'\tLearning rate: {self.optimizer.param_groups[0]["lr"]}')
            print(f'Callbacks: {", ".join(c.__class__.__name__ for c in self.callbacks)}')

        for epoch in range(n_epochs):
            print(f'\nEpoch {epoch}:')
            self.training_epoch(epoch)

            if validation:
                self.validation_epoch(epoch)


class SupervisedTraining(TrainingStrategy):

    def training_epoch(self, epoch: int) -> None:
        self.model.network.train()
        self.emit(TrainingEvents.TRAINING_EPOCH_START, epoch, self.model)

        with tqdm(self.dataset.training_dataloader(), desc='Training', unit='batch') as t:
            for batch_id, batch in enumerate(t):
                self.emit(TrainingEvents.TRAINING_BATCH_START, batch, batch_id, epoch)
                loss, y_pred, y_true = self.model.training_fn(batch, batch_id, epoch)
                self.model.backprop_fn(loss, self.optimizer)
                self.emit(TrainingEvents.TRAINING_BATCH_FINISH, batch, batch_id, epoch,
                          loss.detach(), y_pred.detach(), y_true)

                # update progress bar
                t.set_postfix(self.model.progressbar_metrics())

        self.emit(TrainingEvents.TRAINING_EPOCH_FINISH, epoch, self.model)

    def validation_epoch(self, epoch: int) -> None:
        self.model.network.eval()
        self.model.network.train(False)
        with torch.no_grad():
            self.emit(TrainingEvents.VALIDATION_EPOCH_START, epoch, self.model)

            with tqdm(self.dataset.validation_dataloader(), desc='Validation', unit='batch') as t:
                for batch_id, batch in enumerate(t):
                    self.emit(TrainingEvents.VALIDATION_BATCH_START, batch, batch_id, epoch)
                    loss, y_pred, y_true = self.model.validation_fn(batch, batch_id, epoch)
                    self.emit(TrainingEvents.VALIDATION_BATCH_FINISH, batch, batch_id, epoch,
                              loss.detach(), y_pred.detach(), y_true)

                    # update progress bar
                    t.set_postfix(self.model.progressbar_metrics())

            self.emit(TrainingEvents.VALIDATION_EPOCH_FINISH, epoch, self.model)

    def test(self) -> None:
        self.model.network.eval()
        self.model.network.train(False)
        with torch.no_grad():
            self.emit(TrainingEvents.TESTING_START)

            with tqdm(self.dataset.validation_dataloader(), desc='Testing', unit='batch') as t:
                for batch_id, batch in enumerate(t):
                    self.emit(TrainingEvents.TESTING_BATCH_START, batch, batch_id)
                    loss, y_pred, y_true = self.model.validation_fn(batch, batch_id, -1)
                    self.emit(TrainingEvents.TESTING_BATCH_FINISH, batch, batch_id,
                              loss.detach(), y_pred.detach(), y_true)

                    # update progress bar
                    t.set_postfix(self.model.progressbar_metrics())

            self.emit(TrainingEvents.TESTING_FINISH)



import random
import sys
import subprocess
import platform
import string
from typing import Union
from pathlib import Path
import numpy as np
import torch


def generate_random_id() -> str:
    # https://stackoverflow.com/questions/13484726/safe-enough-8-character-short-unique-random-string
    alphabet = string.ascii_lowercase + string.digits
    return ''.join(random.choices(alphabet, k=8))

def generate_random_seed() -> int:
    return random.randint(-sys.maxsize - 1, sys.maxsize)

def generate_fname(name: str, length_limit: int = 20) -> str:
    # https://stackoverflow.com/questions/295135/turn-a-string-into-a-valid-filename
    alphabet = string.ascii_lowercase + string.digits + '-_'
    return ''.join(c for c in name if c in alphabet)[:length_limit]

def set_seed(seed: int, cudnn_deterministic: bool = False) -> None:
    # https://pytorch.org/docs/master/notes/randomness.html
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

    if cudnn_deterministic and torch.backends.cudnn.enabled:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

def fast_histogram(y_pred: torch.Tensor, y_true: torch.Tensor, num_classes: int) -> torch.Tensor:
    # https://github.com/kevinzakka/pytorch-goodies/blob/master/metrics.py

    mask = (y_true >= 0) & (y_true < num_classes)
    hist = torch.bincount(
        num_classes * y_true[mask] + y_pred[mask],
        minlength=num_classes ** 2,
    ).reshape(num_classes, num_classes).float()
    return hist

def open_folder(path: Union[Path, str]) -> None:
    if platform.system() == 'Windows':
        if isinstance(path, Path):
            subprocess.Popen(f'explorer "{str(path.absolute())}"')
        elif isinstance(path, str):
            subprocess.Popen(f'explorer "{path}"')
    else:
        raise NotImplementedError  # TODO

def open_file(path: Union[Path, str]) -> None:
    if platform.system() == 'Windows':
        if isinstance(path, Path):
            subprocess.Popen(f'explorer /select,"{str(path.absolute())}"')
        elif isinstance(path, str):
            subprocess.Popen(f'explorer /select,"{path}"')
    else:
        raise NotImplementedError  # TODO

def get_current_git_head() -> str:
    # https://stackoverflow.com/questions/14989858/get-the-current-git-hash-in-a-python-script
    return subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip()

def has_nan(x: torch.Tensor) -> bool:
    return torch.isnan(x).any().item()

def to_onehot(tensor: torch.Tensor, num_classes: int) -> torch.Tensor:
    # (N, D, H, W) to (N, num_classes, D, H, W)
    onehot = torch.zeros(tensor.shape[0], num_classes, *tensor.shape[1:],
                         dtype=torch.int32,
                         device=tensor.device)
    return onehot.scatter_(1, tensor.unsqueeze(1), 1)

def readable_number(num: int, unit: str, decimal: int = 3) -> str:
    for mod in ['', 'K', 'M', 'B']:
        if num < 1000:
            break
        num /= 1000
    return f'{num:.{decimal}f} {mod}{unit}'

def readable_time(seconds: int) -> str:
    # https://github.com/pytorch/ignite/blob/master/ignite/_utils.py
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)

    if h > 0:
        return f'{h:.0f}h {m:.0f}m {s:.0f}s'
    elif m > 0:
        return f'{m:.0f}m {s:.1f}s'
    else:
        return f'{s:.1f}s'




import random
import sys
import string
import subprocess
import platform
from typing import Union, List
from pathlib import Path
import numpy as np
import torch
import matplotlib.pyplot as plt


def plot_tensor_5d(data: torch.Tensor, name: str = None, depth_slice: int = None,
                   format: str = 'bcdhw', mode: str = 'column'):
    # assumes B, C, D, H, W

    # TODO oooooooooo
    if format.lower() != 'bcdhw':
        raise NotImplementedError('Only "bcdhw" format is implemented for now')
    if mode.lower() != 'column':
        raise NotImplementedError('Only "column" mode is implemented for now')

    b, c, d, h, w = data.shape
    fig, ax = plt.subplots(nrows=b, ncols=1, sharex=False, sharey=False)

    if name:
        ax[0].set_title(name)

    # if depth_slice is not specified, use middle slice
    if not depth_slice:
        depth_slice = d // 2

    for i in range(b):
        ax[i].imshow(data[i][:, depth_slice, :, :].detach().cpu())

    raise NotImplementedError
    # TODO check if image is shown



from typing import List, Any, Union
import torch
from enum import Enum
import torch.nn.functional as f
import time
from .utils import readable_time
import numpy as np

### Base classes
class Metric:
    def __init__(self):
        self.name = self.__class__.__name__
        self.value = None
        self.visible_progress = True

        self.setup()

    def setup(self) -> None:
        pass

    def set(self, value) -> None:
        self.value = value

    def print(self) -> str:
        try:
            return f'{self.value:.3f}'
        except TypeError:
            return ''

    def on_start(self) -> None:
        pass

    def on_finish(self) -> None:
        pass

    def on_epoch_start(self, epoch: int, model: 'Model') -> None:
        pass

    def on_epoch_finish(self, epoch: int, model: 'Model') -> None:
        pass

    def on_batch_start(self, batch, batch_id: int, epoch: int) -> None:
        pass

    def on_batch_finish(self, batch, batch_id: int, epoch: int,
                        loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        pass

    def on_training_epoch_start(self, epoch: int, model: 'Model') -> None:
        pass

    def on_training_epoch_finish(self, epoch: int, model: 'Model') -> None:
        pass

    def on_validation_epoch_start(self, epoch: int, model: 'Model') -> None:
        pass

    def on_validation_epoch_finish(self, epoch: int, model: 'Model') -> None:
        pass

    def on_testing_start(self) -> None:
        pass

    def on_testing_finish(self) -> None:
        pass

    def on_training_batch_start(self, batch, batch_id: int, epoch: int) -> None:
        pass

    def on_training_batch_finish(self, batch, batch_id: int, epoch: int,
                                 loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        pass

    def on_validation_batch_start(self, batch, batch_id: int, epoch: int) -> None:
        pass

    def on_validation_batch_finish(self, batch, batch_id: int, epoch: int,
                                   loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        pass

    def on_testing_batch_start(self, batch, batch_id: int) -> None:
        pass

    def on_testing_batch_finish(self, batch, batch_id: int,
                                loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        pass

class RunningEpochMetric(Metric):
    def __init__(self):
        super(RunningEpochMetric, self).__init__()

    def on_epoch_start(self, epoch: int, model: 'Model') -> None:
        self.value = None

    def set(self, value) -> None:
        if self.value is None:
            self.value = value
        else:
            self.value = (self.value + value) / 2

### General metrics
class Epoch(Metric):
    def setup(self) -> None:
        super(Epoch, self).setup()
        self.value: int = 0
        self.visible_progress = False

    def on_training_epoch_finish(self, epoch: int, model: 'Model') -> None:
        super(Epoch, self).on_training_epoch_finish(epoch, model)
        self.set(self.value + 1)

class SemSegAccuracy(RunningEpochMetric):
    def on_epoch_start(self, epoch: int, model: 'Model') -> None:
        super(SemSegAccuracy, self).on_epoch_start(epoch, model)
        self.value: float = 0.0

    def on_batch_finish(self, batch, batch_id: int, epoch: int,
                        loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        super(SemSegAccuracy, self).on_batch_finish(batch, batch_id, epoch, loss, y_pred, y_true)

        # convert to sem. seg. map
        y_pred_prob = f.softmax(y_pred, dim=1)
        y_pred_prob = torch.argmax(y_pred_prob, 1)
        accuracy = y_pred_prob.eq(y_true).sum().item() / y_true.nelement()
        del y_pred_prob
        self.set(accuracy)

class Loss(RunningEpochMetric):
    def on_epoch_start(self, epoch: int, model: 'Model') -> None:
        super(Loss, self).on_epoch_start(epoch, model)
        self.value: float = 0.0

    def on_batch_finish(self, batch, batch_id: int, epoch: int,
                        loss: torch.Tensor, y_pred: torch.Tensor, y_true: torch.Tensor) -> None:
        super(Loss, self).on_batch_finish(batch, batch_id, epoch, loss, y_pred, y_true)
        self.set(loss.item())

class EpochTimer(Metric):
    def setup(self) -> None:
        super(EpochTimer, self).setup()
        self.visible_progress = False

    def on_epoch_start(self, epoch: int, model: 'Model') -> None:
        super(EpochTimer, self).on_epoch_start(epoch, model)
        self.value = time.time()

    def on_epoch_finish(self, epoch: int, model: 'Model') -> None:
        super(EpochTimer, self).on_epoch_finish(epoch, model)
        self.value = time.time() - self.value

    def print(self) -> str:
        return readable_time(time.time() - self.value)



# generic dataset classes
# MrcDataset
# EmDataset ?
# ImageDataset
from torch.utils.data import DataLoader

class Dataset:

    def __init__(self):
        pass

    def training_dataloader(self, batch_size: int = 1, shuffle: bool = True) -> DataLoader:
        raise NotImplementedError

    def validation_dataloader(self, batch_size: int = 1, shuffle: bool = True) -> DataLoader:
        raise NotImplementedError

    def test_dataloader(self, batch_size: int = 1, shuffle: bool = True) -> DataLoader:
        raise NotImplementedError

    def __repr__(self):
        return f'Dataset: {self.__class__.__name__}\n'\
               f'\tTraining size: {len(self.training_dataloader())}\n' \
               f'\tValidation size: {len(self.validation_dataloader())}\n' \
               f'\tTest size: {len(self.validation_dataloader())}'

    def print_summary(self) -> None:
        print(repr(self))



from typing import List, Union
from .project import Project
from .utils import *

class Experiment:

    def __init__(self, project: Project, name: str = 'Untitled Experiment', notes: str = '', seed: int = None,
                 devices: Union[List[int], List[str], List[torch.device]] = None, tensorboard: bool = True):

        # general properties
        self.id: str = generate_random_id()
        self.project: Project = project
        self.name: str = name
        self.fname: str = generate_fname(name)
        self.notes: str = notes
        self.path: Path = project.path / self.fname
        self.path.mkdir(parents=True)  # TODO: this should never happen, but it can raise FileExistError

        # save git commit id at the moment of running the experiment
        # TODO bonus points: make .patch file to see what was changed since last commit!
        # TODO bonus bonus points: method to open github website at that revision?
        self.code = get_current_git_head()

        # determinism
        # cudnn_deterministic is slower, but gives almaximum possible determinism
        if seed is None:
            self.seed: int = generate_random_seed()
        else:
            self.seed: int = seed
        set_seed(self.seed, cudnn_deterministic=False)

        # devices
        if devices is None:
            self.devices = [0]
        else:
            self.devices = devices
        # during model initialization torch spills some memory to the GPU0 (not necessarily GPU that we will use)
        torch.cuda.set_device(self.devices[0])

        # tensorboard
        self.tensorboard = tensorboard
        # TODO figure out how (where) to integrate tensorboard - maybe a separate new process to run tensorboard + trainer

        # history
        self.history = []
        # TODO history saving
        # some kind of history rows, to be able to store actions and their results
        # and to store metrics history?

    def new_trainer(self, ):
        pass

    # TODO decorator for functions that should log entry and exit?
    def log(self, action: str,):
        pass



from .project import Project
from .experiment import Experiment
from .model import Model
from .metric import Metric, RunningEpochMetric
from .strategy import TrainingEvents, SupervisedTraining
from .callback import Callback
from .dataset import Dataset

# Exports
__all__ = [
    'Project',
    'Experiment',
    'Model',
    'Metric', 'RunningEpochMetric',
    'TrainingEvents', 'SupervisedTraining',
    'Callback',
    'Dataset',
]



import torch
import torch.nn.functional as F
from torch import nn as nn
from ..core.utils import to_onehot

class TverskyLoss(nn.Module):
    def __init__(self, alpha: float = 0.5, beta: float = 0.5, classes: int = 13):
        super(TverskyLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.n_classes = classes
        self.eps = 1e-7

    def forward(self, y_pred, y_true):

        y_true_onehot = to_onehot(y_true, self.n_classes).float()
        prob = F.softmax(y_pred, dim=1)

        dims = (0, 2, 3, 4)
        intersection = torch.sum(prob * y_true_onehot, dims)
        fps = torch.sum(prob * torch.sub(1, y_true_onehot), dims)
        fns = torch.sum(torch.sub(1, prob) * y_true_onehot, dims)

        del y_true_onehot, prob

        num = intersection
        den = intersection + (self.alpha * fps) + (self.beta * fns)
        tversky_index = (num / (den + self.eps))

        return torch.sub(self.n_classes, tversky_index.sum())



import setuptools
import re

# version fetch
with open('kinlin/__init__.py', 'r') as f:
    version = re.search('__version__ = "(.+?)"', f.read()).group(1)

setuptools.setup(
    name='kinlin',
    version=version,
    description='WIP: PyTorch based framework',  # TODO
    license='MIT',
    author='the-lay',
    author_email='ilja@gubins.lv',
    url='https://github.com/the-lay/kinlin',
    keywords = [],  # TODO
    install_requires = [  # TODO
        'pytorch', 'mrcfile', 'numpy'
    ],
    packages = ['kinlin'],
    test_suite='tests',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
    ]
)



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import math
from .blocks import ConvBnActivation3D, DecoderBlock3D, EncoderBlock3D, init_weights

class UNet(nn.Module):
    def __init__(self, filters: list, in_channels: int = 1, n_classes: int = 13, residual: bool = True, bn: bool = True,
                 activation: str = 'leakyrelu', activation_inplace: bool = True):
        super(UNet, self).__init__()

        self.filters = filters
        self.depth = len(filters)
        if self.depth < 2:
            raise ValueError('Not enough filters, check UNet init code')

        self.in_channels = in_channels
        self.n_classes = n_classes

        self.residual = residual
        self.bn = bn
        self.activation = activation
        self.activation_inplace = activation_inplace

        # down layers
        self.down = nn.ModuleList([EncoderBlock3D(in_channels, filters[0], residual=residual, bn=bn, activation=activation,
                                                  activation_inplace=activation_inplace)])
        for i in range(1, self.depth):
            self.down.append(EncoderBlock3D(filters[i-1], filters[i], residual=residual, bn=bn, activation=activation,
                                            activation_inplace=activation_inplace))

        # bottleneck layers
        self.center = nn.ModuleList([ConvBnActivation3D(filters[self.depth - 1], filters[self.depth - 1], bn=bn, activation=activation,
                                                        activation_inplace=activation_inplace)])

        # up layers
        self.up = nn.ModuleList()
        for i in range(self.depth - 1, 0, -1):
            self.up.append(DecoderBlock3D(filters[i], filters[i], filters[i - 1], bn=bn, activation=activation,
                           activation_inplace=activation_inplace))
        self.up.append(DecoderBlock3D(filters[0], filters[0], self.n_classes, bn=bn, activation=activation,
                                      activation_inplace=activation_inplace))
        self.up.append(nn.Conv3d(self.n_classes, self.n_classes, kernel_size=1, bias=True))

        # init
        init_weights(self, 'kaiming')

    def forward(self, x):

        skips = []
        for i in range(0, self.depth):
            x = self.down[i](x)
            skips.append(x)

        center = x
        for i in range(0, len(self.center)):
            center = self.center[i](center)

        x = center
        for i in range(0, self.depth):
            x, _ = self.up[i](x, skips[self.depth - 1 - i])

        # output layer
        x = self.up[self.depth](x)

        return x



from typing import List
from .utils import *
# from .experiment import Experiment

class Project:
    def __init__(self, name: str = 'Untitled Project', notes: str = '', path: Union[Path, str] = None):

        # general properties
        self.id: str = generate_random_id()
        self.name: str = name
        self.fname: str = generate_fname(name)
        self.notes: str = notes
        if isinstance(path, str):
            path = Path(path)
        self.path: Path = path / self.fname
        self.path.mkdir(parents=True)  # TODO: this should never happen, but it can raise FileExistError

        # experiments
        # self.experiments: List[Experiment] = []

    def new_experiment(self, name: str):

        # new_exp = Experiment(project=self)
        # setup experiment?
        # self.experiments += new_exp
        pass

    def continue_experiment(self):
        pass

    def open_experiment_folder(self):
        open_folder(self.path)




from .core import Project, Experiment, Model, Metric, RunningEpochMetric, TrainingEvents, SupervisedTraining, Callback, Dataset
from .core import utils
from . import metrics
from . import callbacks
from . import networks
from . import datasets
from . import losses


__version__ = "0.0.1"
